# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from socketserver import TCPServer, StreamRequestHandler
from testkitbackend.backend import Backend


class Server(TCPServer):
    allow_reuse_address = True

    def __init__(self, address):
        class Handler(StreamRequestHandler):
            def handle(self):
                backend = Backend(self.rfile, self.wfile)
                try:
                    while backend.process_request():
                        pass
                finally:
                    backend.close()
                print("Disconnected")
        super(Server, self).__init__(address, Handler)
